import asyncio
import hashlib
import io
import json
import os
import re
import traceback

from PyPDF2 import PdfReader

import asyncio

from .search import search_google

from .http import ThrottledClientSession, get_header
from .papers import likely_pdf
from .scrape import browser_scrape


async def search_semantic_scholar(
    query,
    num_results=10,
    semantic_scholar_api_key=None,
    _offset=0,
):
    endpoint = "https://api.semanticscholar.org/graph/v1/paper/search"
    params = {
        "query": query,
        "fields": ",".join(
            [
                "title",
                "year",
                "authors",
                "citationStyles",
                "externalIds",
                "url",
                "openAccessPdf",
                "isOpenAccess",
                # "influentialCitationCount",
                "tldr",
            ]
        ),
        "limit": num_results,
        "offset": _offset,
    }
    ssheader = get_header()
    if semantic_scholar_api_key is not None:
        ssheader["x-api-key"] = semantic_scholar_api_key
    else:
        ssheader["x-api-key"] = os.environ.get("SEMANTIC_SCHOLAR_API_KEY", "")

    async with ThrottledClientSession(
        rate_limit=15 / 60, headers=ssheader
    ) as ss_session:
        async with ss_session.get(url=endpoint, params=params) as response:
            if response.status != 200:
                raise Exception(
                    f"Error searching papers: {response.status} {response.reason} {await response.text()}"
                )
            data = await response.json()
            papers = data.get("data", [])
            # papers.sort(key=lambda x: x["influentialCitationCount"], reverse=True)

    results = []
    for paper in papers[:num_results]:
        title = paper["title"]
        year = paper["year"]
        authors = [author["name"] for author in paper["authors"]]

        bibtex = paper["citationStyles"]["bibtex"]

        # if there is an ArXiv external ID, resolve this as the URL
        if "arXiv" in paper["externalIds"]:
            paper["url"] = f"https://arxiv.org/pdf/{paper['externalIds']['arXiv']}"

        # if openAccessPdf is a URL, use that
        if (
            "openAccessPdf" in paper
            and paper["openAccessPdf"] is not None
            and "url" in paper["openAccessPdf"]
        ):
            paper["url"] = paper["openAccessPdf"]["url"]

        elif "PubMed" in paper["externalIds"]:
            paper["url"] = (
                f"https://pubmed.ncbi.nlm.nih.gov/{paper['externalIds']['PubMed']}"
            )

        elif "PMC" in paper["externalIds"]:
            paper["url"] = (
                f"https://www.ncbi.nlm.nih.gov/pmc/articles/PMC{paper['externalIds']['PMC']}"
            )

        elif "DOI" in paper["externalIds"]:
            paper["url"] = f"https://sci-hub.wf/{paper['externalIds']['DOI']}"
        else:
            # try to find the pdf by name?
            pass

        result = {
            "id": paper["paperId"],
            "title": title,
            "abstract": paper["tldr"],
            "year": year,
            "authors": authors,
            "url": paper["url"],
            "source_type": "semantic_scholar",
        }
        results.append(result)

    return results


async def download_from_semantic_scholar(paper):
    if "id" not in paper:
        paper["id"] = hashlib.sha256(paper["url"].encode()).hexdigest()

    async with ThrottledClientSession(
        rate_limit=15 / 60, headers=get_header()
    ) as session:
        async with session.get(paper["url"], allow_redirects=True) as r:
            try:
                if r.status != 200 or not await likely_pdf(r):
                    raise Exception(f"No paper with id {paper['id']}")

                content = await r.read()  # Read the content
                pdf_file = io.BytesIO(content)  # Create BytesIO object
                pdf = PdfReader(pdf_file)  # Pass BytesIO object to PdfReader
                full_text = "\n".join(page.extract_text() for page in pdf.pages)
                if full_text.strip():  # Check if the text is empty or whitespace
                    paper["full_text"] = full_text
                    return paper
            except Exception as e:
                print(f"Couldn't read PDF the easy way: {e}")

        # if we couldnt get the pdf the easy way, lets try another way
        # if the paper has a title, lets try to google search it but focused on pdf
        results = await search_google(
            '"' + paper["title"] + '"', file_type="pdf", num_results=3
        )
        if len(results) > 0:
            print(f"Found {len(results)} results for {paper['title']}")
            for result in results:
                # try to download the pdf
                try:
                    async with session.get(result["url"], allow_redirects=True) as r:
                        if r.status != 200 or not await likely_pdf(r):
                            raise Exception(f"No paper with id {paper['id']}")
                        content = await r.read()
                        pdf_file = io.BytesIO(content)
                        pdf = PdfReader(pdf_file)
                        full_text = "\n".join(page.extract_text() for page in pdf.pages)
                        if full_text.strip():  # Check if the text is empty or whitespace
                            print("Found full text")
                            # print first 100 characters
                            print(full_text[:100])
                            paper["full_text"] = full_text
                            return paper
                except Exception as e:
                    print(f"Couldn't read PDF from google result: {e}")
                    # stack trace
                    traceback.print_exc()

            # if none of the results have a full text, lets scrape it
            text = await browser_scrape(paper["url"])
            paper["full_text"] = text
            return paper


if __name__ == "__main__":
    query = "Artificial Intelligence"
    num_results = 2

    def run_search_papers(query, num_results=10, semantic_scholar_api_key=None):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            search_semantic_scholar(query, num_results, semantic_scholar_api_key)
        )

    # Search for papers
    search_results = run_search_papers(query, num_results)
    print(f"Search results: {search_results}")

    def run_download_from_semantic_scholar_sync(papers):
        async def run_download_from_semantic_scholar(papers):
            tasks = [download_from_semantic_scholar(paper) for paper in papers]
            return await asyncio.gather(*tasks)
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(run_download_from_semantic_scholar(papers))

    # Download papers
    downloaded_papers = run_download_from_semantic_scholar_sync(search_results)
    # save all downloaded papers to ./test_results
    with open("./test_results.json", "w") as f:
        json.dump(downloaded_papers, f, indent=4, ensure_ascii=False)
        
    assert len(search_results) == num_results
    for result in search_results:
        assert all(key in result for key in ["id", "authors", "url", "year"])

    for result in downloaded_papers:
        assert all(key in result for key in ["id", "authors", "url", "year", "full_text"])

    print("search_semantic_scholar and download_from_semantic_scholar tests passed!")



# sl_sources/__main__.py

import argparse
import asyncio
import json
from . import *

async def main():
    parser = argparse.ArgumentParser(description="Search and download from various sources")
    parser.add_argument("action", choices=["search", "download"], help="Action to perform")
    parser.add_argument("source", choices=["claimminer", "google_scholar", "openalex", "google", "semantic_scholar", "twitter", "youtube"], help="Source to search or download from")
    parser.add_argument("query", help="Search query or ID")
    parser.add_argument("--num_results", type=int, default=10, help="Number of results to return (for search)")
    parser.add_argument("--output", help="Output file for results (optional)")

    args = parser.parse_args()

    if args.action == "search":
        if args.source == "claimminer":
            results = await search_claimminer(args.query, args.num_results)
        elif args.source == "google_scholar":
            results = await search_google_scholar(args.query, args.num_results)
        elif args.source == "openalex":
            results = await search_openalex(args.query, args.num_results)
        elif args.source == "google":
            results = await search_google(args.query, args.num_results)
        elif args.source == "semantic_scholar":
            results = await search_semantic_scholar(args.query, args.num_results)
        elif args.source == "twitter":
            results = await search_twitter(args.query, args.num_results)
        elif args.source == "youtube":
            results = await search_youtube(args.query, args.num_results)
    
    elif args.action == "download":
        if args.source == "google_scholar":
            results = await download_from_google_scholar({"url": args.query})
        elif args.source == "openalex":
            results = await download_from_openalex({"id": args.query})
        elif args.source == "google":
            results = await download_from_google_search({"url": args.query})
        elif args.source == "semantic_scholar":
            results = await download_from_semantic_scholar([{"paperId": args.query}])
        elif args.source == "twitter":
            results = await download_twitter({"tweet_id": args.query})
        elif args.source == "youtube":
            results = await download_from_youtube({"url": args.query})
        else:
            print(f"Download not supported for {args.source}")
            return

    if args.output:
        with open(args.output, 'w') as f:
            json.dump(results, f, indent=2)
    else:
        print(json.dumps(results, indent=2))

if __name__ == "__main__":
    asyncio.run(main())


import re
from typing import Dict, Any
import requests
from bs4 import BeautifulSoup

async def get_paper_details(url: str) -> Dict[str, Any]:
    if "arxiv.org" in url:
        return await get_arxiv_details(url)
    elif "pubmed.ncbi.nlm.nih.gov" in url:
        return await get_pubmed_details(url)
    else:
        raise ValueError("Unsupported URL")

async def get_arxiv_details(url: str) -> Dict[str, Any]:
    arxiv_id = re.search(r'(?:arxiv\.org/abs/)(.+)', url).group(1)
    api_url = f'http://export.arxiv.org/api/query?id_list={arxiv_id}'
    
    async with requests.get(api_url) as response:
        soup = BeautifulSoup(await response.text, 'lxml-xml')
        
        entry = soup.find('entry')
        title = entry.find('title').text
        authors = [author.find('name').text for author in entry.find_all('author')]
        abstract = entry.find('summary').text
        
        return {
            "id": arxiv_id,
            "title": title,
            "authors": authors,
            "abstract": abstract
        }

async def get_pubmed_details(url: str) -> Dict[str, Any]:
    pubmed_id = re.search(r'(?:pubmed\.ncbi\.nlm\.nih\.gov/)(.+)', url).group(1)
    api_url = f'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi?db=pubmed&id={pubmed_id}&retmode=json'
    
    async with requests.get(api_url) as response:
        data = await response.json()
        
        result = data['result'][pubmed_id]
        title = result['title']
        authors = result['authors']
        abstract = result['summary']
        
        return {
            "id": pubmed_id,
            "title": title,
            "authors": [f'{author["name"]}' for author in authors],
            "abstract": abstract
        }
    
async def likely_pdf(response):
    try:
        text = await response.text()
        text = text.lower()
        if "invalid article id" in text:
            return False
        if "no paper" in text or "not found" in text:
            return False
        if "404" in text and "error" in text:
            return False
        if "404" in text and "not found" in text:
            return False
        if "403" in text and "forbidden" in text:
            return False
    # Bytestream? Probably a PDF
    except UnicodeDecodeError:
        return True
    # we're still unsure, so let's check mimetype
    if response.headers['Content-Type'] == 'application/pdf' or response.content.startswith(b'%PDF-'):
        return True
    return False


async def pubmed_to_pdf_url(url, session):
    # url = f"https://pubmed.ncbi.nlm.nih.gov/{pubmed_id}/"
    pubmed_id = url.split("/")[-1]

    async with session.get(url) as r:
        if r.status != 200:
            raise Exception(
                f"Error fetching PMC ID for PubMed ID {pubmed_id}. {r.status}"
            )
        html_text = await r.text()
        pmc_id_match = re.search(r"PMC\d+", html_text)
        if pmc_id_match is None:
            raise Exception(f"No PMC ID found for PubMed ID {pubmed_id}.")
        pmc_id = pmc_id_match.group(0)
    pmc_id = pmc_id[3:]

    async with session.get(url) as r:
        if r.status != 200:
            raise Exception(f"No paper with pmc id {pmc_id}. {url} {r.status}")
        html_text = await r.text()
        pdf_url = re.search(r'href="(.*\.pdf)"', html_text)
        if pdf_url is None:
            raise Exception(f"No PDF link found for pmc id {pmc_id}. {url}")
        return f"https://www.ncbi.nlm.nih.gov{pdf_url.group(1)}"




import asyncio
import os
from typing import Any, Dict, List

from dotenv import load_dotenv
from twikit import Client

load_dotenv()  # Load environment variables from .env file


async def search_twitter(query: str, num_results: int = 10) -> List[Dict[str, Any]]:
    USERNAME = os.getenv("TWITTER_USERNAME")
    EMAIL = os.getenv("TWITTER_EMAIL")
    PASSWORD = os.getenv("TWITTER_PASSWORD")
    client = Client("en-US")
    await client.login(auth_info_1=USERNAME, auth_info_2=EMAIL, password=PASSWORD)
    tweets = await client.search_tweet(query, "Latest")

    results = []
    for tweet in tweets[:num_results]:
        result = {
            "title": f"Tweet by {tweet.user.name}",
            "url": f"https://twitter.com/{tweet.user.screen_name}/status/{tweet.id}",
            "full_text": tweet.text,
            "created_at": tweet.created_at,
            "tweet_id": tweet.id,
            "source_type": "twitter",
        }
        results.append(result)

    return results


async def download_twitter(tweet_result: Dict[str, Any]) -> Dict[str, Any]:
    USERNAME = os.getenv("TWITTER_USERNAME")
    EMAIL = os.getenv("TWITTER_EMAIL")
    PASSWORD = os.getenv("TWITTER_PASSWORD")
    client = Client("en-US")
    await client.login(auth_info_1=USERNAME, auth_info_2=EMAIL, password=PASSWORD)

    tweet_id = tweet_result["tweet_id"]
    tweet = await client.get_tweet_by_id(tweet_id)

    full_text = tweet.text

    # Get replies to the tweet
    replies = await client.search_tweet(f"to:{tweet.user.screen_name}", "Latest")
    for reply in replies:
        # Check if the reply has a 'in_reply_to_status_id' attribute
        if (
            hasattr(reply, "in_reply_to_status_id")
            and reply.in_reply_to_status_id == tweet_id
        ):
            full_text += f"\n\nReply by {reply.user.name}: {reply.text}"

    # Check if the tweet is a reply to another tweet
    if hasattr(tweet, "in_reply_to_status_id") and tweet.in_reply_to_status_id:
        try:
            parent_tweet = await client.get_tweet_by_id(tweet.in_reply_to_status_id)
            full_text = f"In reply to {parent_tweet.user.name}: {parent_tweet.text}\n\n{full_text}"
        except Exception as e:
            print(f"Error fetching parent tweet: {e}")

    # Check if the tweet is a retweet
    if hasattr(tweet, "retweeted_status") and tweet.retweeted_status:
        try:
            retweeted_tweet = await client.get_tweet_by_id(tweet.retweeted_status.id)
            full_text = f"Retweet of {retweeted_tweet.user.name}: {retweeted_tweet.text}\n\n{full_text}"
        except Exception as e:
            print(f"Error fetching retweeted tweet: {e}")

    result = {
        "title": tweet_result["title"],
        "url": tweet_result["url"],
        "full_text": full_text,
        "created_at": tweet_result["created_at"],
        "source_type": "twitter",
    }

    return result


if __name__ == "__main__":

    async def test_search_twitter():
        query = "Python programming"
        num_results = 3
        results = await search_twitter(query, num_results)

        assert len(results) == num_results

        for result in results:
            assert "title" in result
            assert "url" in result
            assert "full_text" in result
            assert "created_at" in result

        print("search_twitter test passed!")

    async def test_download_twitter():
        query = "Python programming"
        num_results = 1
        search_results = await search_twitter(query, num_results)

        for search_result in search_results:
            downloaded_result = await download_twitter(search_result)

            assert "title" in downloaded_result
            assert "url" in downloaded_result
            assert "full_text" in downloaded_result
            assert "created_at" in downloaded_result

        print("download_twitter test passed!")

    asyncio.run(test_search_twitter())
    asyncio.run(test_download_twitter())



from .claimminer import search_claimminer
from .search import download_from_google_search, search_google
from .google_scholar import download_from_google_scholar, search_google_scholar
from .openalex import download_from_openalex, search_openalex
from .semantic_scholar import download_from_semantic_scholar, search_semantic_scholar
from .twitter import search_twitter
from .youtube import download_from_youtube, search_youtube

CHUNK_SIZE = 5


async def search_source(source, query, num_results):
    if source == "google":
        print("Searching Google")
        results = await search_google(query, num_results=num_results)
        print(f"Found {len(results)} results from Google")
        return results
    elif source == "semantic_scholar":  # arxiv, pubmed, sagepub
        print("Searching Semantic Scholar")
        results = await search_semantic_scholar(query, num_results=num_results)
        print(f"Found {len(results)} results from Semantic Scholar")
        return results
    elif source == "twitter":
        print("Searching Twitter")
        results = await search_twitter(query, num_results)
        print(f"Found {len(results)} results from Twitter")
        return results
    elif source == "youtube":
        print("Searching Youtube")
        results = await search_youtube(query, num_results=num_results)
        print(f"Found {len(results)} results from Youtube")
        return results
    elif source == "google_scholar":
        print("Searching Google Scholar")
        results = await search_google_scholar(query, num_results=num_results)
        print(f"Found {len(results)} results from Google Scholar")
        return results
    elif source == "openalex":
        print("Searching OpenAlex")
        results = await search_openalex(query, num_results=num_results)
        print(f"Found {len(results)} results from OpenAlex")
        return results
    elif source == "claimminer":
        claimminer_results = await search_claimminer(query, cutoff=0.7)
        print(f"Found {len(claimminer_results)} results from ClaimMiner")
        return claimminer_results
    else:
        return []


async def download_source(search_result):
    print(f"Downloading {search_result['url']}")
    source_type = search_result.get("source_type", "crawl")
    if source_type == "crawl":
        # scrapes the url directly
        result = await download_from_google_search(search_result)
    elif source_type == "google":
        result = await download_from_google_search(search_result)
    elif source_type == "twitter":
        result = search_result
    elif source_type == "youtube":
        result = await download_from_youtube(search_result)
    elif source_type == "google_scholar":
        result = await download_from_google_scholar(search_result)
    elif source_type == "semantic_scholar":
        result = await download_from_semantic_scholar(search_result)
    elif source_type == "openalex":
        result = await download_from_openalex(search_result)
    else:
        result = search_result

    return result



import ssl

ssl_context = ssl.create_default_context()
ssl_context.check_hostname = False
ssl_context.verify_mode = ssl.CERT_NONE

import asyncio
import hashlib
import io
import json
import traceback
from typing import Any, Dict, List

import aiohttp
from aiohttp import ClientConnectorCertificateError, ClientError
from pyalex import Works
from PyPDF2 import PdfReader

from .papers import likely_pdf

from .http import ThrottledClientSession, get_header
from .scrape import browser_scrape
from .search import search_google

def reconstruct_abstract(inverted_index):
    if not inverted_index:
        return None
    
    words = []
    max_position = max(max(positions) for positions in inverted_index.values())
    
    for i in range(max_position + 1):
        for word, positions in inverted_index.items():
            if i in positions:
                words.append(word)
                break
    
    return ' '.join(words)


async def search_openalex(query: str, num_results: int) -> List[Dict[str, Any]]:
    results = []

    try:
        works = (
            Works()
            .search(query)
            .select(["id", "doi", "title", "publication_year", "authorships", "open_access", "abstract_inverted_index", "ids"])
            .paginate(per_page=num_results)
        )

        for page in works:
            for work in page:
                authors = []
                for authorship in work.get("authorships", []):
                    author = authorship.get("author", {}).get("display_name")
                    if author:
                        authors.append(author)
                authors_str = ", ".join(authors)

                result = {
                    "id": work.get("id"),
                    "doi": work.get("doi"),
                    "authors": authors,
                    "title": work.get("title"),
                    "year": work.get("publication_year"),
                    "url": None,
                    "source_type": "openalex",
                }

                # Resolve URL based on available identifiers
                if work.get("open_access", {}).get("oa_status") == "gold" and work.get("open_access", {}).get("oa_url"):
                    result["url"] = work["open_access"]["oa_url"]
                elif work.get("ids", {}).get("arxiv"):
                    result["url"] = f"https://arxiv.org/pdf/{work['ids']['arxiv']}.pdf"
                elif work.get("ids", {}).get("pmid"):
                    result["url"] = f"https://pubmed.ncbi.nlm.nih.gov/{work['ids']['pmid']}"
                elif work.get("ids", {}).get("pmcid"):
                    result["url"] = f"https://www.ncbi.nlm.nih.gov/pmc/articles/{work['ids']['pmcid']}"

                if result["url"] is None:
                        
                    # Google search for PDF or any other file type
                    try:
                        print(f"attempting google search for pdf: {work['title']} {authors_str}")

                        search_results = await search_google(f'"{work["title"]}" {authors_str}', file_type="pdf", num_results=5)
                        if not search_results or len(search_results) == 0:
                            
                            print(f"no pdf found, searching for html: {work['title']} by {authors_str}")
                            search_results = await search_google(f'"{work["title"]}" {authors_str}', num_results=5)

                        async with ThrottledClientSession(rate_limit=15 / 60, headers=get_header()) as session:
                            print(f"looking for pdf...")
                            for search_result in search_results:
                                try:
                                    async with session.get(search_result["url"], allow_redirects=True) as r:
                                        if r.status == 200 and await likely_pdf(r):
                                            result["url"] = search_result["url"]
                                            break
                                except ClientConnectorCertificateError:
                                    print(f"SSL certificate verification failed for {search_result['url']}")
                                except Exception as e:
                                    print(f"Error checking URL {search_result['url']}: {e}")
                                await asyncio.sleep(1)  # Add a delay between requests

                        if not result["url"] and search_results:
                            result["url"] = search_results[0]["url"]
                    except Exception as e:
                        print(f"Error during Google search for {work['title']}: {e}")

                if result["url"] is None and work.get("doi"):
                    result["url"] = f"https://doi.org/{work['doi']}"

                # Try to get the abstract
                try:
                    abstract_inverted_index = work.get("abstract_inverted_index")
                    if abstract_inverted_index is None:
                        print(f"Debug: No abstract_inverted_index for work {work.get('id')}")
                        # Try to get the abstract directly if available
                        result["abstract"] = work.get("abstract")
                    else:
                        result["abstract"] = reconstruct_abstract(abstract_inverted_index)

                except Exception as e:
                    print(f"Error fetching full work details for {work['id']}: {e}")
                    result["abstract"] = None

                # if the url is still none, lets do a google search for the pdf by name and authors (joined)
                if result["url"] is None:
                    print(f"no url found, searching for any type: {work['title']}")
                    authors_str = ", ".join(authors)
                    search_results = await search_google(f'"{work["title"]}" {authors_str}', file_type="pdf", num_results=5)
                    try:
                        async with ThrottledClientSession(rate_limit=15 / 60, headers=get_header()) as session:
                        # see if its a pdf
                            for search_result in search_results:    
                                async with session.get(search_result["url"], allow_redirects=True) as r:
                                    if r.status == 200 and await likely_pdf(r):
                                        result["url"] = search_result["url"]
                                        break
                                
                            if result["url"] is None and len(search_results) > 0:
                                result["url"] = search_results[0]["url"]
                    except Exception as e:
                        print(f"Error during Google search for {work['title']}: {e}")
                        traceback.print_exc()

                # if the url is still none, lets do a google search, not sorted by pdf
                if result["url"] is None:
                    print(f"no url found, searching for any type: {work['title']}")
                    search_results = await search_google(f'"{work["title"]}" {authors_str}', num_results=5)
                    if search_results:
                        result["url"] = search_results[0]["url"]

                # okay, now we're really in trouble, just search by title, this is extreme edge case
                if result["url"] is None:
                    search_results = await search_google(f'"{work["title"]}"', num_results=5)
                    if search_results and len(search_results) > 0:
                        result["url"] = search_results[0]["url"]

                results.append(result)

            if len(results) >= num_results:
                break

    except Exception as e:
        print(f"Error in search_openalex: {e}")
        traceback.print_exc()

    # Save search results to a JSON file
    try:
        with open("openalex_search_results.json", "w") as file:
            json.dump(results, file, indent=4)
    except Exception as e:
        print(f"Error saving search results to JSON: {e}")

    return results[:num_results]

async def download_and_extract_text(url: str, session: aiohttp.ClientSession) -> str:
    async with session.get(url, allow_redirects=True) as response:
        if response.status == 200:
            content_type = response.headers.get("Content-Type", "").lower()
            if "application/pdf" in content_type:
                pdf_content = await response.read()
                pdf_file = io.BytesIO(pdf_content)
                try:
                    pdf_reader = PdfReader(pdf_file)
                    text = "\n".join(page.extract_text() for page in pdf_reader.pages)
                    return text
                except Exception as e:
                    print(f"Error extracting text from PDF: {e}")
            else:
                return await browser_scrape(url)
        else:
            raise Exception(f"Failed to download content from {url}. Status code: {response.status}")


async def download_from_openalex(work: Dict[str, Any]) -> Dict[str, Any]:
    if "id" not in work:
        work["id"] = hashlib.sha256(work["title"].encode()).hexdigest()

    async with ThrottledClientSession(
        rate_limit=15 / 60, headers=get_header()
    ) as session:
        try:
            work["full_text"] = await download_and_extract_text(work["url"], session)
        except Exception as e:
            print(f"Unexpected error downloading full text: {e}")
            traceback.print_exc()

        if not work.get("full_text"):
            # Search for the PDF on Google
            search_results = await search_google(f'"{work["title"]}"', file_type="pdf", num_results=3)
            for search_result in search_results:
                try:
                    text = await download_and_extract_text(search_result["url"], session)
                    if text:
                        work["full_text"] = text
                        print(f"Downloaded full text from Google search: {search_result['url']}")
                        return work
                except Exception as e:
                    print(f"Error downloading PDF from Google search: {e}")
                    print("Trying next...")

            # Scrape the original URL
            try:
                text = await browser_scrape(work["url"])
                if text:
                    work["full_text"] = text
                    print(f"Downloaded full text from original URL: {work['url']}")
                    return work
            except Exception as e:
                print(f"Error scraping full text: {e}")
                print("Trying google...")
                search_results = await search_google(f'"{work["title"]}"', num_results=3)
                url = search_results[0]["url"]
                text = await download_and_extract_text(url, session)
                if text:
                    work["full_text"] = text
                    print(f"Downloaded full text from Google search: {url}")
                    return work

    return work
async def main():
    query = "Artificial Intelligence"
    num_results = 5

    search_results = await search_openalex(query, num_results)
    print(f"Search results: {len(search_results)}")

    download_tasks = [download_from_openalex(result) for result in search_results]
    downloaded_results = await asyncio.gather(*download_tasks)

    # Save downloaded results to a JSON file
    with open("openalex_downloaded_results.json", "w") as file:
        json.dump(downloaded_results, file, indent=4)


if __name__ == "__main__":
    asyncio.run(main())


import os
import sys
import json
import requests
import copy
import aiohttp
import asyncio

search_api_url = "https://claimminer.societylibrary.org/api/search"


def get_claimminer_api_token():
    # Fetch the password from environment variable
    password = os.environ.get("CLAIMMINER_PASSWORD")
    username = os.environ.get("CLAIMMINER_USERNAME")

    if not password:
        raise ValueError("Please set the CLAIMMINER_PASSWORD environment variable")

    # URL to get the access token
    token_url = "https://claimminer.societylibrary.org/api/token"

    # Data for the token request
    data = {"grant_type": "", "username": username, "password": password}

    # Request to get the access token
    response = requests.post(token_url, data=data)
    response_data = response.json()
    access_token = response_data.get("access_token")

    if not access_token:
        raise ValueError("Failed to retrieve access token")

    return access_token


# URLs & Endpoints
# Retrieve Access Token
try:
    access_token = get_claimminer_api_token()
    claimminer_available = True
except ValueError as e:
    print(f"Warning: {e}. ClaimMiner functionality will be disabled.")
    access_token = None
    claimminer_available = False

if access_token:
    headers = {
        "Accept": "application/json",
        "Authorization": f"Bearer {access_token}",
    }
else:
    headers = {
        "Accept": "application/json",
    }


def prepend_text(data):
    # Prepend question text
    data["question"] = "[question] " + data["question"]

    for position in data["positions"]:
        # Prepend position text
        position["position"] = "[position] " + position["position"]

        for category in position["categories"]:
            # Prepend category text
            category["category"] = "[category] " + category["category"]

            for claim in category["claims"]:
                # Prepend claim text
                claim["claim"] = "[including reasons like] " + claim["claim"]

                for i in range(len(claim["arguments"])):
                    # Prepend argument text
                    claim["arguments"][i] = (
                        "[to be more specific this means] " + claim["arguments"][i]
                    )


async def get_document_url_async(doc_id, session):
    if not claimminer_available:
        print("ClaimMiner is not available.")
        return None

    document_url = f"https://claimminer.societylibrary.org/api/document/{doc_id}"
    async with session.get(document_url, headers=headers) as response:
        if response.ok:
            document_info = await response.json()
            return document_info["url"]
        else:
            print(f"Error fetching document URL: {response.status} - {response.reason}")
            return None


# Function to search each claim and print the claim along with the top result
async def search_claimminer(text, cutoff=0.7):

    if not claimminer_available:
        print("ClaimMiner is not available.")
        return []

    form_data = {
        "search_text": text,
        "offset": 0,
        "limit": 10,
        "mode": "semantic",
        "model": "all_minilm_l6_v2",
        "lam": 0.7,
        "include_paragraphs": "true",
        "include_statements": "false",
        "one_per_doc": "false",
        "one_per_cluster": "false",
        "group_by_cluster": "false",
        "only_with_quote": "false",
        "include_sentences": "true",
    }

    async with aiohttp.ClientSession() as session:
        async with session.post(
            search_api_url, headers=headers, data=form_data
        ) as response:
            structured_results = []

            if response.ok:
                results = await response.json()

                if results:
                    top_result = results[0]
                    top_result_text = top_result["result"]["text"]
                    top_result_rank = top_result["rank"]
                    # print(f"Claim: {text}\nTop Result: {top_result_text}\nRank: {top_result_rank}\n")
                else:
                    print(f"No results found for the claim: {text}")

                for result in results:
                    top_result_rank = result["rank"]
                    if top_result_rank > cutoff:
                        doc_id = result["result"]["doc_id"]
                        document_url = await get_document_url_async(doc_id, session)
                        # print(document_url)
                        # print(result['result']['text'])
                        # print(text)
                        if document_url:
                            result_object = {
                                "quote": result["result"]["text"],
                                "url": document_url,
                                "similarity_score": top_result_rank,
                                "index": "claimminer",
                                "source_type": "claimminer",  # added to track source in bifocal browser
                                "full_text": result["result"][
                                    "text"
                                ],  # added full text to result object to be like other sources
                            }
                            structured_results.append(result_object)
            else:
                print(
                    f"Error searching the text: {response.status} - {response.reason}"
                )

    return structured_results


async def process_directory(debate_maps_folder="debate maps/documentary/todo"):

    # Loop through each file in the given directory
    for file in os.listdir(debate_maps_folder):
        if file.endswith(".json"):
            json_file_path = os.path.join(debate_maps_folder, file)
            print(f"Processing: {json_file_path}")

            # Load the JSON file
            with open(json_file_path, "r") as json_file:
                debate_graph = json.load(json_file)

            # Apply processing to each debate graph
            for index, position_obj in enumerate(debate_graph["positions"]):
                position = position_obj["position"]
                search_results = await search_claimminer(position)
                print(search_results)
                sys.exit()
                reference_urls_set = set()
                if search_results:
                    for result in search_results:
                        if "url" in result:
                            reference_urls_set.add(result["url"])
                reference_urls = list(reference_urls_set)
                if reference_urls:
                    debate_graph["positions"][index]["reference_urls"] = reference_urls

            # prepend_text(debate_graph)

            # Prepare and save the updated JSON file
            new_file_name_parts = file.split(".json")
            new_file_name = f"{new_file_name_parts[0]} (with references).json"
            new_file_path = os.path.join(debate_maps_folder, new_file_name)

            with open(new_file_path, "w") as new_json_file:
                json.dump(debate_graph, new_json_file, indent=4)

            print(f"Updated debate graph saved to '{new_file_name}'")

            # Make a deep copy of the debate_graph (so original is unchanged)
            debate_graph_no_refs = copy.deepcopy(debate_graph)

            # Visit each position and remove 'reference_urls'
            for position_obj in debate_graph_no_refs["positions"]:
                if "reference_urls" in position_obj:
                    del position_obj["reference_urls"]

            # Specify the path for the "no_reference" folder or create it if it doesn't exist
            no_ref_folder = os.path.join(debate_maps_folder, "no_reference")
            if not os.path.exists(no_ref_folder):
                os.makedirs(no_ref_folder)

            # Prepare the filename with the suffix "(with labels)" for saving without reference URLs
            original_file_name_without_extension = os.path.splitext(file)[0]
            no_ref_file_name = (
                f"{original_file_name_without_extension} (with labels).json"
            )
            no_ref_file_path = os.path.join(no_ref_folder, no_ref_file_name)

            # Save the modified copy without reference URLs but with the new suffix
            with open(no_ref_file_path, "w") as no_ref_json_file:
                json.dump(debate_graph_no_refs, no_ref_json_file, indent=4)

            print(f"Debate graph saved to '{no_ref_file_path}'")


if __name__ == "__main__":
    default_claim = (
        "Artificial intelligence will revolutionize the healthcare industry."
    )
    claim_input = input(
        f"Enter a claim to search in ClaimMiner (or press Enter to use the default claim '{default_claim}'): "
    )

    if not claim_input.strip():
        claim_input = default_claim

    async def search_and_print_results(claim):
        search_results = await search_claimminer(claim)
        print(f"Search results for the claim: {claim}")
        for result in search_results:
            print(f"Quote: {result['quote']}")
            print(f"URL: {result['url']}")
            print(f"Similarity Score: {result['similarity_score']}")
            print("---")

    asyncio.run(search_and_print_results(claim_input))

    """

    asyncio.run(process_directory())

    for index, position_obj in enumerate(debatege_graph['positions']):
        position = position_obj['position']
        reference_urls = search_text_and_print_top_result(position)
        if reference_urls:
            debate_graph['positions'][index]['reference_urls'] = reference_urls

    # Creating a new filename based on the existing one
    new_file_name_parts = debate_graph_file_name.split('.json')
    new_file_name = f"{new_file_name_parts[0]} (with references).json"

    prepend_text(debate_graph)

    # Save the updated `debate_graph` back to a new JSON file
    with open(new_file_name, 'w') as new_file:
        json.dump(debate_graph, new_file, indent=4)

    print(f"Updated debate graph saved to '{new_file_name}'")

    sys.exit()

    # Iterate through each position and category to extract claims
    for position in debate_graph['positions']:
        for category in position['categories']:
            for claim_obj in category['claims']:
                claim = claim_obj['claim']
                search_text_and_print_top_result(claim)

    """



import asyncio
import hashlib
import json
import os
import traceback
from typing import List

from dotenv import load_dotenv

from .scrape import browser_scrape
import aiohttp

load_dotenv()  # Load environment variables from .env file

GOOGLE_SEARCH_API_KEY = os.getenv("GOOGLE_API_KEY")
CSE_ID = os.getenv("GOOGLE_CSE_ID")


async def search_google(query, num_results=10, file_type=None):
    results = []
    google_api_url = "https://customsearch.googleapis.com/customsearch/v1"

    params = {
        "key": GOOGLE_SEARCH_API_KEY,
        "cx": CSE_ID,
        "q": query,
        "num": num_results,
    }

    if file_type:
        params["fileType"] = file_type

    async with aiohttp.ClientSession() as session:
        async with session.get(google_api_url, params=params) as response:
            response_json = await response.json()

    if "items" in response_json:
        for item in response_json["items"]:
            title = item.get("title")
            abstract = item.get("snippet")
            url = item.get("link")
            # hash an id from the url
            id = hashlib.md5(url.encode()).hexdigest()
            results.append(
                {
                    "id": id,
                    "title": title,
                    "url": url,
                    "abstract": abstract,
                    "source_type": "google",
                }
            )
    return results



async def download_from_google_search(search_result: List[str]) -> List[str]:
    try:
        search_result["full_text"] = await browser_scrape(search_result["url"])
        return search_result
    except Exception as e:
        print(f"Error scraping {search_result['url']}: {e}")
        print(traceback.format_exc())
        return search_result
    
if __name__ == "__main__":
    query = "Artificial Intelligence"
    num_results = 2

    def search_google_sync():
        # get the loop
        loop = asyncio.get_event_loop()
        # run the search_google function
        search_results = loop.run_until_complete(search_google(query, num_results))
        return search_results

    # Search for results
    search_results = search_google_sync()
    print(f"Search results: {search_results}")

    with open("search_results_google.json", "w") as f:
        json.dump(search_results, f, indent=4, ensure_ascii=False)

    async def run_download_from_google_search(search_results):
        tasks = [download_from_google_search(result) for result in search_results]
        return await asyncio.gather(*tasks)

    # Download pages
    downloaded_pages = asyncio.run(run_download_from_google_search(search_results))
    
    # Save all downloaded pages to ./test_results_google.json
    with open("./download_results_google.json", "w") as f:
        json.dump(downloaded_pages, f, indent=4, ensure_ascii=False)
        
    assert len(search_results) == num_results
    for result in search_results:
        assert all(key in result for key in ["id", "title", "url", "abstract", "source_type"])

    for result in downloaded_pages:
        assert all(key in result for key in ["id", "title", "url", "abstract", "source_type", "full_text"])

    print("search_google and download_from_google_search tests passed!")


# sl_sources/__init__.py

from .claimminer import search_claimminer
from .google_scholar import search_google_scholar, download_from_google_scholar
from .openalex import search_openalex, download_from_openalex
from .search import search_google, download_from_google_search
from .semantic_scholar import search_semantic_scholar, download_from_semantic_scholar
from .twitter import search_twitter, download_twitter
from .youtube import search_youtube, download_from_youtube
from .sources import search_source, download_source
from .scrape import check_for_playwright, get_browser_and_page, browser_scrape
from .crawl import crawl

__all__ = [
    "search_source",
    "download_source",
    "search_claimminer",
    "search_google_scholar",
    "download_from_google_scholar",
    "search_openalex",
    "download_from_openalex",
    "search_google",
    "download_from_google_search",
    "search_semantic_scholar",
    "download_from_semantic_scholar",
    "search_twitter",
    "download_twitter",
    "search_youtube",
    "download_from_youtube",
    "browser_scrape",
    "check_for_playwright",
    "get_browser_and_page",
    "crawl",
]



import asyncio
import hashlib
import io
import json
import re
from typing import Any, Dict, List
from PyPDF2 import PdfReader

from .search import search_google
from .papers import likely_pdf, pubmed_to_pdf_url

from .http import ThrottledClientSession, get_header
from .scrape import browser_scrape, get_browser_and_page

async def search_google_scholar(
    query: str, num_results: int = 10
) -> List[Dict[str, Any]]:
    results = []
    from playwright.async_api import async_playwright

    async with async_playwright() as p:
        # set the browser path
        browser, page = await get_browser_and_page(p)

        search_url = f"https://scholar.google.com/scholar?q={query}&num={num_results}"
        try:
            await page.goto(search_url, wait_until="networkidle", timeout=20000)
        except Exception as e:
            # probably timeout and page went ok
            print(f"Error navigating to {search_url}: {str(e)}")
        
        # wait for 3 seconds
        await asyncio.sleep(3)

        for item in await page.query_selector_all(".gs_r.gs_or.gs_scl"):
            title_element = await item.query_selector(".gs_rt a")
            title = await title_element.inner_text() if title_element else ""
            url = await title_element.get_attribute("href") if title_element else ""

            authors_year_element = await item.query_selector(".gs_a")
            authors_year_text = await authors_year_element.inner_text() if authors_year_element else ""
            authors, publication, year = parse_authors_year(authors_year_text)

            snippet_element = await item.query_selector(".gs_rs")
            snippet = await snippet_element.inner_text() if snippet_element else ""
            # set id from hash of the title
            id = hashlib.sha256(title.encode()).hexdigest()

            result = {
                "id": id,
                "title": title,
                "authors": authors,
                "publication": publication,
                "year": year,
                "abstract": snippet,
                "url": url,
                "source_type": "google_scholar",
            }
            results.append(result)

        await browser.close()

        return results[:num_results]

def clean_text(text):
    text = re.sub(r'\u2026', '', text)  # Remove the ellipsis character
    text = re.sub(r'\s*-\s*', '', text)  # Remove any " - "
    text = re.sub(r'\s+', ' ', text)  # Replace multiple spaces with a single space
    return text.strip()

def parse_authors_year(authors_year_text):
    # Try to extract the year using a regular expression
    year_match = re.search(r'\b(19|20)\d{2}\b', authors_year_text)
    if year_match:
        year = year_match.group()
        authors_publication = authors_year_text.replace(year, '').strip()
    else:
        year = ""
        authors_publication = authors_year_text.strip()
    
    # Split authors and publication
    parts = authors_publication.split("-")
    authors = parts[0].strip()
    publication = "-".join(parts[1:]).strip() if len(parts) > 1 else ""

    # Split authors into an array
    authors_array = [clean_text(author) for author in re.split(r',\s+|\s+and\s+', authors)]
    
    # Clean publication and year
    publication = clean_text(publication)
    year = clean_text(year)

    return authors_array, publication, year


async def download_from_google_scholar(item: Dict[str, Any]) -> Dict[str, Any]:
    async with ThrottledClientSession(rate_limit=15 / 60, headers=get_header()) as session:
        # Check for external identifiers
        if "arxiv" in item["url"]:
            item["url"] = f"https://arxiv.org/pdf/{item['url'].split('/')[-1]}"
        elif "pubmed.ncbi.nlm.nih.gov" in item["url"]:
            item["url"] = await pubmed_to_pdf_url(item["url"], session)
        elif "ncbi.nlm.nih.gov/pmc" in item["url"]:
            pmc_id = item["url"].split("/")[-1]
            item["url"] = f"https://www.ncbi.nlm.nih.gov/pmc/articles/PMC{pmc_id}/pdf/"
        elif "doi.org" in item["url"]:
            # if item has a doi, use sci-hub to get the pdf
            item["url"] = f"https://sci-hub.wf/{item['url']}"
        try:
            async with session.get(item["url"], allow_redirects=True) as r:
                if r.status == 200 and await likely_pdf(r):
                    content = await r.read()
                    pdf_file = io.BytesIO(content)
                    pdf = PdfReader(pdf_file)
                    full_text = "\n".join(page.extract_text() for page in pdf.pages)
                    if full_text.strip():
                        item["full_text"] = full_text
                        return item
        except Exception as e:
            print(f"Couldn't read PDF: {e}")

            # Try searching for the PDF on Google
            results = await search_google('"' + item["title"] + '"', file_type="pdf", num_results=5)
            if len(results) > 0:
                for result in results:
                    # if result contains books.google, skip
                    if "books.google.com" in result["url"]:
                        continue
                    async with session.get(result["url"], allow_redirects=True) as r:
                        if r.status == 200 and await likely_pdf(r):
                            try:
                                content = await r.read()
                                pdf_file = io.BytesIO(content)
                                pdf = PdfReader(pdf_file)
                                full_text = "\n".join(page.extract_text() for page in pdf.pages)
                                if full_text.strip():
                                    item["full_text"] = full_text
                                    return item
                            except Exception as e:
                                print(f"Couldn't read PDF: {e}")
            else:
                item["full_text"] = await browser_scrape(item["url"])
                return item

        # Scrape the HTML
        try:
            item["full_text"] = await browser_scrape(item["url"])
            return item
        except Exception as e:
            print(f"Couldn't scrape HTML: {e}")

    full_text = "Error downloading or scraping"
    item["full_text"] = full_text
    print ("******** item['url'] is")
    print (item["url"])
    return item

if __name__ == "__main__":

    async def test_main():
        # Test the search function
        query = "Artificial Intelligence"
        num_results = 5
        search_results = await search_google_scholar(query, num_results)

        assert len(search_results) == num_results
        for result in search_results:
            assert all(
                key in result
                for key in ["title", "authors", "year", "url"]
            )

        print("search_google_scholar test passed!")

        # Test the download function asynchronously
        download_tasks = [download_from_google_scholar(result) for result in search_results]
        downloaded_results = await asyncio.gather(*download_tasks)

        # write all of the results to a file
        with open("google_scholar_results.json", "w") as file:
            json.dump(downloaded_results, file)
            
        for downloaded_result in downloaded_results:
            assert (
                len(downloaded_result.get("full_text", "").split()) > 100
            ), "Transcript seems too short"
            print(f"download_from_google_scholar test passed!")

    # start test_main
    asyncio.run(test_main())


import asyncio
import glob
import json
import os
import re
import tempfile
import urllib.parse
from multiprocessing import Pool
from typing import Any, Dict, List

import requests
from bs4 import BeautifulSoup
from yt_dlp import YoutubeDL
import librosa


def fetch_youtube_data(url) -> str:
    try:
        response = requests.get(url=url)
        response.raise_for_status()
    except requests.RequestException:
        raise requests.RequestException("Failed to fetch data from YouTube.")
    return response.text


def search_by_url(url: str):
    if "https://" in url:
        response = fetch_youtube_data(url)
        soup_obj = BeautifulSoup(response, features="lxml")
        video_id = re.search(r"(?<=\?v=)[\w-]+", url).group(0)
        title = soup_obj.find("meta", {"name": "title"})["content"]
        js_script = str(soup_obj.find_all("script")[20])
        duration_mil = re.search(r'"approxDurationMs":"(\d+)"', js_script).group(1)
        description = soup_obj.find("meta", {"name": "description"})["content"]
        if description is None:
            # find the ytd-text-inline-expander" elemenent and get all the text
            description = soup_obj.find("ytd-text-inline-expander").get_text()


        return {
            "id": video_id,
            "title": title,
            "url": url,
            "duration": duration_mil,
            "long_desc": description,
        }
    else:
        raise ValueError("Please provide valid URL.")


def search_by_term(term: str, max_results: int = None):
    encoded_search = urllib.parse.quote_plus(term)
    BASE_URL = "https://youtube.com"
    url = f"{BASE_URL}/results?search_query={encoded_search}&sp=CAM"
    response = fetch_youtube_data(url)

    results = []
    searched_obj = _prepare_data(response)
    for contents in searched_obj:
        for video in contents["itemSectionRenderer"]["contents"]:
            res = {}
            if "videoRenderer" in video.keys():
                video_data = video.get("videoRenderer", {})
                res["id"] = video_data.get("videoId", None)
                res["thumbnails"] = [
                    thumb.get("url", None)
                    for thumb in video_data.get("thumbnail", {}).get("thumbnails", [{}])
                ]
                res["title"] = (
                    video_data.get("title", {}).get("runs", [[{}]])[0].get("text", None)
                )
                res["long_desc"] = (
                    video_data.get("descriptionSnippet", {})
                    .get("runs", [{}])[0]
                    .get("text", None)
                )
                
                res["channel"] = (
                    video_data.get("longBylineText", {})
                    .get("runs", [[{}]])[0]
                    .get("text", None)
                )
                res["duration"] = video_data.get("lengthText", {}).get("simpleText", 0)
                res["views"] = video_data.get("viewCountText", {}).get("simpleText", 0)
                res["publish_time"] = video_data.get("publishedTimeText", {}).get(
                    "simpleText", 0
                )
                res["url_suffix"] = (
                    video_data.get("navigationEndpoint", {})
                    .get("commandMetadata", {})
                    .get("webCommandMetadata", {})
                    .get("url", None)
                )
                results.append(res)

        if results:
            if max_results is not None and len(results) > max_results:
                return results[:max_results]
        break
    return results


def _prepare_data(response):
    start = response.index("ytInitialData") + len("ytInitialData") + 3
    end = response.index("};", start) + 1
    json_str = response[start:end]
    data = json.loads(json_str)
    searched_obj = data["contents"]["twoColumnSearchResultsRenderer"][
        "primaryContents"
    ]["sectionListRenderer"]["contents"]

    return searched_obj


async def search_youtube(query: str, num_results: int) -> List[Dict[str, Any]]:
    search_results = search_by_term(query, max_results=num_results)

    results = []
    for search_result in search_results:
        video_id = search_result["id"]
        video_url = f"https://www.youtube.com/watch?v={video_id}"
        video_data = search_by_url(video_url)

        result = {
            "id": video_id,
            "title": search_result["title"],
            "abstract": search_result["long_desc"] or video_data["long_desc"],
            "year": search_result["publish_time"][
                :4
            ],  # Extract the year from publish_time
            "authors": [search_result["channel"]],
            "url": video_url,
            "source_type": "youtube",
        }
        results.append(result)

    return results


def download_video(url, temp_dir):
    ydl_opts = {
        "overwrites": True,
        "format": "bestaudio",
        "outtmpl": os.path.join(temp_dir, "audio.mp3"),
    }
    with YoutubeDL(ydl_opts) as ydl:
        ydl.download(url)
        return os.path.join(temp_dir, "audio.mp3")


def format_timestamp(
    seconds: float, always_include_hours: bool = False, decimal_marker: str = "."
):
    if seconds is not None:
        milliseconds = round(seconds * 1000.0)
        hours = milliseconds // 3_600_000
        milliseconds -= hours * 3_600_000
        minutes = milliseconds // 60_000
        milliseconds -= minutes * 60_000
        seconds = milliseconds // 1_000
        milliseconds -= seconds * 1_000
        hours_marker = f"{hours:02d}:" if always_include_hours or hours > 0 else ""
        return f"{hours_marker}{minutes:02d}:{seconds:02d}{decimal_marker}{milliseconds:03d}"
    else:
        return seconds


from tqdm import tqdm


def transcribe(audio_file):
    from transformers import WhisperProcessor, WhisperForConditionalGeneration

    processor = WhisperProcessor.from_pretrained("openai/whisper-tiny.en")
    model = WhisperForConditionalGeneration.from_pretrained("openai/whisper-tiny.en")

    # Load the audio file using librosa
    audio_data, sampling_rate = librosa.load(audio_file)

    # Resample the audio to 16000 Hz
    resampled_audio = librosa.resample(
        audio_data, orig_sr=sampling_rate, target_sr=16000
    )

    # Define chunk size (in seconds)
    chunk_length_s = 30
    chunk_length_samples = chunk_length_s * 16000  # 16000 is the target sampling rate

    # Split audio into chunks
    audio_chunks = [
        resampled_audio[i : i + chunk_length_samples]
        for i in range(0, len(resampled_audio), chunk_length_samples)
    ]

    transcription = []

    # Process each chunk
    for chunk in tqdm(audio_chunks, desc="Transcribing"):
        inputs = processor(chunk, sampling_rate=16000, return_tensors="pt")

        if "input_features" in inputs:
            input_features = inputs["input_features"]
        else:
            raise ValueError("The processor output does not contain 'input_features'.")

        outputs = model.generate(
            input_features,
            output_scores=False,
            return_dict_in_generate=True,
            output_attentions=False,
        )

        for sequence in outputs.sequences:
            chunk_text = processor.decode(sequence, skip_special_tokens=True)
            transcription.append(chunk_text)

    full_transcription = " ".join(transcription)
    return full_transcription


def download_from_youtube_sync(search_result) -> Dict[str, Any]:
    print(f"Downloading video from youtube: {search_result['url']}")

    with tempfile.TemporaryDirectory() as temp_dir:
        ydl_opts = {
            "writesubtitles": True,
            "subtitleslangs": ["en"],
            "subtitlesformat": "vtt",
            "outtmpl": os.path.join(temp_dir, "%(title)s.%(ext)s"),
            "skip_download": True,
            "quiet": False,
        }

        with YoutubeDL(ydl_opts) as ydl:
            ydl.download([search_result["url"]])

        vtt_files = glob.glob(os.path.join(temp_dir, "*.vtt"))
        if vtt_files:
            vtt_file = vtt_files[0]
            try:
                with open(vtt_file, "r", encoding="utf-8") as f:
                    vtt_data = f.read()
                transcript = vtt_to_text(vtt_data)
                search_result["full_text"] = transcript
                return search_result
            except Exception as e:
                print(f"Error reading VTT file: {e}")
        else:
            print("No VTT file found, attempting to transcribe audio with Whisper")
            try:
                id = search_result["url"].split("v=")[1].split("&")[0]
                audio_file = os.path.join(temp_dir, f"{id}")

                # def duration_filter(info, *, incomplete):
                #     duration = info.get("duration")
                #     if duration:
                #         if duration > 6000:  # 5400 seconds = 100 minutes
                #             return "The video is longer than 100 minutes"
                #     return None

                ydl_opts = {
                    "format": "bestaudio/best",
                    "outtmpl": audio_file,
                    "postprocessors": [
                        {
                            "key": "FFmpegExtractAudio",
                            "preferredcodec": "mp3",
                            "preferredquality": "192",
                        }
                    ],
                    # "match_filter": duration_filter,
                }

                with YoutubeDL(ydl_opts) as ydl:
                    ydl.download([search_result["url"]])

                if os.path.exists(audio_file + ".mp3"):
                    transcript = transcribe(audio_file + ".mp3")
                    search_result["full_text"] = transcript
                    return search_result
                else:
                    print(f"Error: Audio file {audio_file}.mp3 not found.")
                    search_result["full_text"] = None
                    return search_result
            except Exception as e:
                print(f"Error in audio transcription: {e}")
                search_result["full_text"] = None
                return search_result


async def download_from_youtube(search_result) -> Dict[str, Any]:
    # run async on a separate thread so it doesn't block the main thread
    loop = asyncio.get_running_loop()
    with Pool(processes=1) as pool:
        result = await loop.run_in_executor(
            None, pool.apply, download_from_youtube_sync, (search_result,)
        )
    return result


def vtt_to_text(vtt_data: str) -> str:
    print("Converting VTT to text")
    lines = vtt_data.strip().split("\n")
    transcript = []

    for line in lines[2:]:  # Skip the first two lines (WEBVTT and blank line)
        if "-->" not in line and not line.strip().isdigit():
            transcript.append(line.strip())

    return " ".join(transcript)


async def test_search_youtube():
    query = "Python tutorial"
    num_results = 2
    results = await search_youtube(query, num_results)

    assert len(results) == num_results

    for result in results:
        assert "title" in result
        assert "url" in result
        assert "description" in result

    print("Test passed!")


async def test_download_and_transcribe():
    # Test video without captions
    video_url_no_captions = {"url": "https://www.youtube.com/watch?v=xuCn8ux2gbs"}
    transcript_no_captions = await download_from_youtube(video_url_no_captions)
    assert transcript_no_captions, "Failed to transcribe video"
    print("Successfully transcribed video")

    # Add assertions to check the content of the transcripts
    assert (
        len(transcript_no_captions.get("full_text", "").split()) > 100
    ), "Transcript seems too short"


async def download_videos_parallel(search_results):
    tasks = [download_from_youtube(result) for result in search_results]
    gathered = await asyncio.gather(*tasks)
    return gathered


async def main():
    # Test parallel downloading
    query = "self introduction"
    num_results = 3
    search_results = await search_youtube(query, num_results)
    # save search results to youtube_search.json
    with open("youtube_search.json", "w") as f:
        json.dump(search_results, f)

    download_results = await download_videos_parallel(search_results)

    # save download results
    with open("youtube_download.json", "w") as f:
        json.dump(download_results, f)

    print("Parallel download and transcription test passed!")


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import hashlib
import traceback
import aiohttp
import json
import os

from .papers import get_paper_details
from .sources import download_source, search_source

async def cloud_function_request(request_type, params):
    print("cloud_function_request")
    print(request_type)
    print(params)
    async with aiohttp.ClientSession() as session:
        async with session.post(
            os.getenv('CLOUD_FUNCTION_URL'),
            json={"request_type": request_type, **params}
        ) as response:
            if response.status == 200:
                return await response.json()
            else:
                print(f"Error calling cloud function: {response.status}")
                return None
            
async def evaluate_page(text, research_topic, model="gpt-4o-mini"):
    # Load OpenAI API key from environment variable
    OPENAI_API_KEY = os.getenv('OPENAI_API_KEY', os.getenv('SOCIETY_API_KEY'))
    if not OPENAI_API_KEY:
        raise ValueError("Please set the OPENAI_API_KEY environment variable")
    
    # Truncate text if it's too long (adjust max_length as needed)
    max_length = 128000
    if len(text) > max_length:
        text = text[:max_length] + "..."

    # Prepare the prompt for GPT-4
    prompt = f"""
    Analyze the following text and determine if it's relevant. 
    Also extract any URLs mentioned in the text that seem relevant to these topics.
    
    Text:
    {text}
    
    We are researching the following topic and related domains:
    {research_topic}

    Please evaluate if the text above contains relevant and substantive information for our research.

    Respond with a JSON object containing two fields:
    1. "relevant": a boolean indicating if the text is relevant
    2. "links": a list of relevant URLs extracted from the text which are worth looking at. Ignore links that are not relevant to the research topic.
    3. "summary": a summary of the text, focusing on the most relevant information for the research topic.
    
    Example response:
    {{
        "relevant": true,
        "links": ["https://example.com/ai-article", "https://example.org/ml-study"],
        "summary": "A summary of the text, focusing on the most relevant information for the research topic."
    }}
    """

    # Make API call to OpenAI
    async with aiohttp.ClientSession() as session:
        async with session.post(
            "https://api.openai.com/v1/chat/completions",
            headers={
                "Authorization": f"Bearer {OPENAI_API_KEY}",
                "Content-Type": "application/json"
            },
            json={
                "model": model,
                "messages": [{"role": "user", "content": prompt}],
                "temperature": 0.2
            }
        ) as response:
            if response.status == 200:
                result = await response.json()                
                try:
                    evaluation = result['choices'][0]['message']['content']
                    # Remove the ```json and ``` if present
                    evaluation = evaluation.replace('```json\n', '').replace('\n```', '')
                    evaluation = json.loads(evaluation)
                    evaluation['text'] = text
                    return evaluation
                except json.JSONDecodeError:
                    print(f"Error parsing GPT-4 response: {result['choices'][0]['message']['content']}")
                    return {"relevant": False, "links": [], "summary": "", "text": text}
            else:
                print(f"Error calling OpenAI API: {response.status}")
                return {"relevant": False, "links": [], "summary": "", "text": text}

async def crawl(keywords=[], urls=[], sources=["google_scholar", "openalex"], research_topic="", max_depth=3, use_cloud_function=False):
    urls = set(urls) # make sure urls are unique
    keywords = set(keywords) # make sure keywords are unique

    # resolve arxiv and pubmed
    arxiv_urls = [url for url in urls if "arxiv.org" in url]
    pubmed_urls = [url for url in urls if "pubmed.ncbi.nlm.nih.gov" in url]
    urls = [url for url in urls if "arxiv.org" not in url and "pubmed.ncbi.nlm.nih.gov" not in url]

    youtube_urls = [url for url in urls if "youtube.com" in url] # move any url that contains youtube.com to youtube_urls
    twitter_urls = [url for url in urls if "twitter.com" in url or "x.com" in url] # same for twitter

    # filter urls that are not youtube or twitter
    urls = [url for url in urls if "youtube.com" not in url and "twitter.com" not in url and "x.com" not in url]

    # # create a results array for each type
    # direct_sources = [{"url": url, "source_type": "crawl"} for url in urls]
    # youtube_sources = [{"url": url, "source_type": "youtube"} for url in youtube_urls]
    # twitter_sources = [{"url": url, "source_type": "twitter"} for url in twitter_urls]

    search_tasks = []
    num_results = 3

    for source in sources:
        for keyword in keywords:
            print(f"Searching for {keyword} with source {source}")
            if use_cloud_function:
                task = asyncio.create_task(cloud_function_request("search", {"source_type": source, "query": keyword, "num_results": num_results}))
            else:
                task = asyncio.create_task(search_source(source, keyword, num_results=num_results))
            search_tasks.append(task)

    search_results = await asyncio.gather(*search_tasks)
    # flatten search_results into a single list
    search_results = [item for sublist in search_results for item in sublist]

    # join direct_sources, youtube_sources, twitter_sources, and search_results
    # sources = direct_sources + youtube_sources + twitter_sources + search_results
    sources = search_results

    cache = {}
    visited = set()

    async def crawl_links(links, depth, parent_source=None):
        source_type = parent_source['source_type'] if parent_source else "crawl"
        if depth > max_depth:
            return
        tasks = []
        for link in links:
            if link not in visited:
                visited.add(link)
                try:
                    work = await get_paper_details(link)
                except Exception as e:
                    # hash the id from the url
                    id = hashlib.md5(link.encode()).hexdigest()
                    work = {
                        "id": id,
                        "title": "Unknown",
                        "authors": ["Unknown"],
                        "abstract": "Crawled from " + link,
                        "url": link,
                        "source_type": source_type
                    }
                    if parent_source is not None:
                        work['title'] = "Unknown (Crawled from " + parent_source['title'] + ")"
                        work['abstract'] = "Crawled from " + link + " with parent " + parent_source['title'] + " | " + parent_source['url']
                    
                task = asyncio.create_task(download_and_evaluate(work, depth + 1))
                tasks.append(task)
        await asyncio.gather(*tasks)

    if os.path.exists('manifest.json'):
        with open('manifest.json', 'r') as f: 
            cache = json.load(f)

    async def download_and_evaluate(source, depth):
        url = source['url']
        if depth > max_depth:
            return

        url_hash = hashlib.md5(url.encode()).hexdigest()
        if url_hash in cache:
            # crawl_links the links in the cache
            await crawl_links(cache[url_hash]["links"], depth + 1, parent_source=source)
            print(f"Using cached content for {url}")
            return

        try:
            if use_cloud_function:
                # Use cloud function for download
                content = await cloud_function_request("download", {"search_result": source})
                print("content")
                print(content)
            else:
                content = await download_source(source)

            text = content['full_text']

            result = await evaluate_page(text, research_topic)
            if result["relevant"]:
                file_name = f"{url_hash}.txt"
                file_path = os.path.join("./downloaded_data", file_name)
                
                os.makedirs("./downloaded_data", exist_ok=True)
                with open(file_path, 'w', encoding='utf-8') as f:
                    f.write(text)
                cache[url_hash] = {
                    "url": url,
                    "file_path": file_path,
                    "links": result["links"],
                    "summary": result["summary"],
                    "relevant": result["relevant"]
                }
                
                with open('manifest.json', 'w') as f:
                    json.dump(cache, f, indent=2)
                
                await crawl_links(result["links"], depth + 1, parent_source=source)
            else:
                cache[url_hash] = {
                    "url": url,
                    "links": result["links"],
                    "summary": result["summary"],
                    "relevant": result["relevant"]
                }
        except Exception as e:
            print(f"Error crawling {url}: {e}")
            # print the exception location
            print(traceback.format_exc())

    tasks = []
    for source in sources:
        visited.add(source['url'])
        task = asyncio.create_task(download_and_evaluate(source, 0))
        tasks.append(task)

    await asyncio.gather(*tasks)

async def main():
    research_topic = "can humans keep up? the impact of artificial intelligence on neuroscience"
    keywords = ["artificial intelligence", "machine learning", "neuroscience and the ai race"]
    urls = []

    sources = [
        'google',
        # 'semantic_scholar',
        # 'google_scholar',
        # 'openalex',
        # 'twitter', # TODO: needs work
        # 'youtube',
    ]
    
    use_cloud_function = os.getenv('CLOUD_FUNCTION_ENABLED', 'false').lower() == 'true'
    await crawl(keywords, urls, sources, research_topic=research_topic, max_depth=2, use_cloud_function=use_cloud_function)
    
    print("Crawl completed. Check the 'downloaded_data' directory and 'manifest.json' for results.")

if __name__ == "__main__":
    asyncio.run(main())



import os
from io import BytesIO
from pathlib import Path
import random
import sys
import time

import aiohttp
from bs4 import BeautifulSoup, NavigableString
from PyPDF2 import PdfReader

import os
from pathlib import Path
import sys
import asyncio

from typing import Optional

link_blacklist = [
    "home",
    "next",
    "about us",
    "contact",
    "log in",
    "account",
    "sign",
    "sign up",
    "sign in",
    "sign out",
    "privacy",
    "close",
    "privacy policy",
    "terms of service",
    "terms and conditions",
    "terms",
    "conditions",
    "privacy",
    "legal",
    "guidelines",
    "filter",
    "theme",
    "english",
    "accessibility",
    "authenticate",
    "join",
    "edition",
    "subscribe",
    "news",
    "home",
    "blog",
    "jump to",
    "español",
    "world",
    "europe",
    "politics",
    "profile",
    "election",
    "health",
    "business",
    "tech",
    "sports",
    "advertise",
    "advertising",
    "ad",
    "banner",
    "sponsor",
    "promotion",
    "promoted",
]


element_blacklist = [
    "sidebar",
    "nav",
    "footer",
    "header",
    "menu",
    "account",
    "login",
    "form",
    "search",
    "advertisement",
    "masthead",
    "popup",
    "overlay",
    "floater",
    "modal",
    "noscript",
    "iframe",
    "script",
    "style",
    "head",
    "meta"
]

def check_for_playwright():
    if sys.platform.startswith("win"):
        browsers_path = Path(os.getenv("LOCALAPPDATA", "")) / "ms-playwright"
    elif sys.platform == "darwin":
        browsers_path = Path.home() / "Library" / "Caches" / "ms-playwright"
    else:  # Linux and other Unix-like OSes
        browsers_path = Path.home() / ".cache" / "ms-playwright"
    # Search for any folder that contains the word 'chromium'
    chromium_folders = list(browsers_path.glob("*chromium*"))

    if chromium_folders:
        for folder in chromium_folders:
            print("Found chromium folder: ", folder)
        chromium_installed = True
    else:
        print("No chromium folder found")
        chromium_installed = False

    if not chromium_installed:
        print("Browser binaries not found. Installing now...")
        # run playwright install chromium in subprocess
        import subprocess

        result = subprocess.run(
            ["playwright", "install", "chromium"], capture_output=True, text=True
        )
        print("Installation output:")
        print(result.stdout)
        if result.stderr:
            print("Error output:")
            print(result.stderr)
    else:
        print("Browser binaries are already installed.")
    # get the browser path from chromium_folders
    chromium_folders = list(browsers_path.glob("*chromium*"))
    browser_path = chromium_folders[0]
    # if on mac, fine the browser path from the browser_path
    if sys.platform == "darwin":
        browser_path = browser_path / "chrome-mac" / "Chromium.app" / "Contents" / "MacOS" / "Chromium"
    elif sys.platform == "linux":
        browser_path = browser_path / "chrome-linux" / "chrome"
    elif sys.platform == "win32":
        browser_path = browser_path / "chrome-win" / "chrome.exe"
    return browser_path

async def get_browser_and_page(playwright, capsolver_api_key=None):
    browser_path = check_for_playwright()

    from playwright_stealth import stealth_async
    # check if CAPSOLVER_API_KEY exists
    if capsolver_api_key is None:
        capsolver_api_key = os.getenv("CAPSOLVER_API_KEY")
    
    if capsolver_api_key is None or capsolver_api_key == "":
        print("Warning: No capsolver API key provided. Attempting to launch browser without a captcha solver")

    from capsolver_extension_python import Capsolver
    extension_path = Capsolver(api_key=capsolver_api_key).load(with_command_line_option=False) 
    user_agent = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36"
    browser = await playwright.chromium.launch(headless=True, executable_path=browser_path, args=[
            '--disable-extensions-except='+ extension_path,
            '--load-extension=' + extension_path,
            '--user-agent=' + user_agent,
        ])
    page = await browser.new_page()
    await stealth_async(page)
    return browser, page

async def extract_plain_text(page):
    content = await page.content()
    soup = BeautifulSoup(content, 'html.parser')
    
    # Remove scripts and style elements
    for element in soup(['script', 'style']):
        element.decompose()

    # Define blacklists
    link_blacklist = ['unwanted_link_keyword1', 'unwanted_link_keyword2']  # Adjust as needed

    # Remove blacklisted links
    for link in soup.find_all('a'):
        href = link.get('href', '').lower()
        text = link.text.lower()
        if any(keyword in href or keyword in text for keyword in link_blacklist):
            link.decompose()

    # Define inline and block elements
    inline_elements = {'span', 'a', 'button', 'strong', 'em', 'i', 'b', 'small', 'code'}
    block_elements = {'div', 'main', 'section', 'article', 'section', 'p', 'li', 'ol', 'ul', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'blockquote'}
    
    # Function to recursively process elements
    def process_element(element):
        if isinstance(element, NavigableString):  # Check if the element is a NavigableString
            return element.strip()
        
        if element.name in inline_elements:
            return ' ' + ' '.join(process_element(child) for child in element.children if child.name != 'br').strip() + ' '
        elif element.name in block_elements:
            return '\n' + '\n'.join(process_element(child) for child in element.children).strip() + '\n'
        else:
            return ' '.join(process_element(child) for child in element.children).strip()

    text = process_element(soup.body or soup)  # Start processing from body or root
    text = ' '.join(text.split())  # Normalize whitespace
    text = text.replace(' \n ', '\n').replace('\n ', '\n').replace(' \n', '\n').strip()  # Final cleanup
    
    return text


def resolve_with_wayback_machine(link: str) -> str:
    from waybackpy import Url
    user_agent = "Mozilla/5.0 (Windows NT 5.1; rv:40.0) Gecko/20100101 Firefox/40.0"
    wayback_url = Url(link, user_agent)
    wayback_url.newest()
    url = wayback_url.archive_url
    return url

wayback_resolution_whitelist = [
    "nytimes",
    "wsj",
    "nyt",
    "latimes",
    "washingtonpost",
    "bbc",
    "cnn",
    "foxnews",
    "nbcnews",
    "cnbc",
    "reuters",
    "apnews",
    "bloomberg",
    "businessinsider",
    "businesswire",
    "theverge"
]

async def browser_scrape(link: str) -> str:
    """
    Asynchronously scrape a source to extract text.
    
    Args:
    link (str): The URL of the webpage to scrape.

    Returns:
    str: The extracted text content from the webpage.
    """
    from playwright.async_api import async_playwright
    async with async_playwright() as p:
        try:
            browser, page = await get_browser_and_page(p)
            if any(keyword in link for keyword in wayback_resolution_whitelist):
                new_link = resolve_with_wayback_machine(link)
                link = new_link if new_link else link
            await page.goto(link, wait_until="networkidle")
            # wait for 2 seconds to defeat any captchas etc
            await asyncio.sleep(2)
            extracted_text = await extract_plain_text(page)
            return extracted_text
        except Exception as e:
            print(f"Error during scraping: {e}")
            # if we've already wayback machine'd it, don't try again
            if any(keyword in link for keyword in wayback_resolution_whitelist):
                return ""
            # try to resolve with wayback machine
            new_link = resolve_with_wayback_machine(link)
            if new_link and new_link != link:
                return await browser_scrape(new_link)
            else:
                return ""
        finally:
            await browser.close()


if __name__ == "__main__":
    import asyncio
    async def run():
        print("Scraping...")
        text = await browser_scrape("https://www.nytimes.com/2024/04/29/technology/ai-google-microsoft.html")
        print("Scraped")
        print(text)

    asyncio.run(run())